using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;


namespace HIDTest
{
	public partial class HIDTestForm : Form
	{
		public const int RESEND_COUNT = 256;

		//Application global variables 
		HIDDevice Device = new HIDDevice();

		// Vendor and product IDs to look for.
		const uint VendorID = 0x04D8;
		const uint ProductID = 0x4541;

		public HIDTestForm()
		{
			InitializeComponent();

			System.Guid HidGuid = System.Guid.NewGuid();
			HIDDevice.HidD_GetHidGuid(ref HidGuid);
			Device.RegisterForDeviceNotifications(this.Handle, HidGuid);
		}

		[DllImport("winmm.dll")]
		static public extern int timeGetTime();

		protected override void WndProc(ref Message m)
		{
			if (m.Msg == HIDDevice.WM_DEVICECHANGE)
			{
				int wParam = m.WParam.ToInt32();

				if (wParam == HIDDevice.DBT_DEVICEARRIVAL)
				{
					int devType = Marshal.ReadInt32(m.LParam, 4);
					if (devType == HIDDevice.DBT_DEVTYP_DEVICEINTERFACE)
					{
						HIDDevice.DEV_BROADCAST_DEVICEINTERFACE devIF;
						devIF = (HIDDevice.DEV_BROADCAST_DEVICEINTERFACE)Marshal.PtrToStructure(m.LParam, typeof(HIDDevice.DEV_BROADCAST_DEVICEINTERFACE));
						// TBD: See if it is the device we are interested in
						// For now, set DeviceDetected to false on any
						// interface change.  Slight performance penalty,
						// but will ensure we don't try talking to a device
						// that isn't there.
						// Ensure the device is available.
						Device.DeviceDetected = false;
					}
				}
				else if (wParam == HIDDevice.DBT_DEVICEREMOVECOMPLETE)
				{
					int devType = Marshal.ReadInt32(m.LParam, 4);
					// A device was removed
					if (devType == HIDDevice.DBT_DEVTYP_DEVICEINTERFACE)
					{
						Device.DeviceDetected = false;
					}
				}
			}
			base.WndProc(ref m);
		}

		int failCount = 0;

		void FlushPendingInputReports()
		{
			bool bResult = false;
			uint nBytesRead = 0;
			uint totalFlushed = 0;
			do
			{
				Win32Comm.ReadFile(Device.ReadHandle, Device.InputReport,
					(uint)Device.Caps.InputReportByteLength, out nBytesRead,
					ref Device.HIDOverlapped);
				// Give the device a little time to respond
				int hRes = Win32Event.WaitForSingleObject(Device.HIDOverlapped.hEvent, 100);

				// Reset event for next use
				Win32Event.ResetEvent(Device.HIDOverlapped.hEvent);

				if (hRes == Win32Event.WAIT_OBJECT_0)
				{
					// Successful read
					bResult = true;
					totalFlushed += nBytesRead;
				}
				else
				{
					bResult = false;
				}
			} while (bResult && (nBytesRead > 0));

			if (totalFlushed > 0)
			{
				string outStr = String.Format("Flushed {0} bytes", totalFlushed);
				DisplayData(outStr);
			}
		}


		bool GetReportBytes()
		{
			bool result = false;
			int hRes;
			uint bytesRead = 0;

			//Read a report
			Win32Comm.ReadFile(Device.ReadHandle, Device.InputReport,
				(uint)Device.Caps.InputReportByteLength, out bytesRead,
				ref Device.HIDOverlapped);
			// Give the device a little time to respond
			hRes = Win32Event.WaitForSingleObject(Device.HIDOverlapped.hEvent, 100);

			if (hRes == Win32Event.WAIT_OBJECT_0)
			{
				// Successful read
				result = true;
			}
			else if (hRes == Win32Event.WAIT_TIMEOUT)
			{
				// Failed to read within the alloted time.
				Win32Comm.CancelIo(Device.ReadHandle);
				Device.CloseHidDevice(true);
			}
			else
			{
				// Unknown error
				Win32Comm.CancelIo(Device.ReadHandle);
				Device.CloseHidDevice(true);
			}

			// Reset event for next use
			Win32Event.ResetEvent(Device.HIDOverlapped.hEvent);

			return result;
		}

		bool WriteReport()
		{
			//Send a report to the device.

			uint BytesWritten = 0;
			string strBytesWritten = "";

			//The first byte is the report number.
			Device.OutputReport[0] = 0;

			//Get the bytes to send from the combo boxes.
			string outText = textBox1.Text;
			int maxLen, outLen;
			if (Device.Caps.InputReportByteLength < Device.Caps.OutputReportByteLength)
				maxLen = Device.Caps.InputReportByteLength;
			else
				maxLen = Device.Caps.OutputReportByteLength;
			outLen = outText.Length;
			if (outLen > maxLen)
				outLen = maxLen;

			//Get the values from the combo boxes.
			for (int bcnt = 1; bcnt < outLen; bcnt++)
				Device.OutputReport[bcnt] = (byte)outText[bcnt - 1];

			/*
			API Function: WriteFile
			Sends a report to the device.
			Returns: success or failure.
			Requires:
			The device handle returned by CreateFile.
			The Output Report length returned by HidP_GetCaps,
			A report to send.
			*/

			bool Result = Win32Comm.WriteFile(Device.HidDevice, Device.OutputReport,
				(uint)Device.Caps.OutputReportByteLength, out BytesWritten, IntPtr.Zero);

			if (!Result)
			{
				//The WriteFile failed, so close the handle, display a message,
				//and set DeviceDetected to false so the next attempt will look for the device.
				Device.CloseHidDevice(true);
				DisplayData("Can't write to device");
			}

			//Display the result of the API call and the report bytes.
			strBytesWritten = String.Format("Bytes Written: {0}", BytesWritten);
			DisplayData(strBytesWritten);

			return Device.DeviceDetected;
		}

		// Send an array of bytes to the device.  They are send in chunks of
		// either len or the length of the output report, whichever is smaller.
		bool SendReportBytes(byte reportID, byte[] buf, int len, bool outFlag, bool inFlag)
		{
			HIDDevice.HidD_FlushQueue(Device.HidDevice);
			bool Result = false;
			int outLen = Device.Caps.OutputReportByteLength - 1; // Subtract 1 for report ID
			int numSends = (len + outLen - 1) / outLen;
			int bytesLeft = len;
			int offset = 0;
			for (int i = 0; i < numSends; i++, offset += outLen)
			{
				uint bytesWritten;
				int bytesToSend;
				if (bytesLeft < outLen)
					bytesToSend = bytesLeft;
				else
					bytesToSend = outLen;
				Array.Clear(Device.OutputReport, 0, outLen);

				Device.OutputReport[0] = reportID; // Report ID

				if (outFlag)
				{
					for (int j = 0; j < bytesToSend; j++)
						Device.OutputReport[j + 1] = buf[offset + j];

					Result = Win32Comm.WriteFile(Device.HidDevice, Device.OutputReport,
					(uint)Device.Caps.OutputReportByteLength, out bytesWritten, IntPtr.Zero);

					if (!Result)
					{
						//The WriteFile failed, so close the handle, display a message,
						//and set DeviceDetected to false so the next attempt will look for the device.
						Device.CloseHidDevice(true);
						DisplayData("Write failed in SendBytes()");
						return false;
					}
				}

				if (inFlag)
				{
					// Read the bytes back
					Result = GetReportBytes();

					if (!Result)
					{
						DisplayData("Can't read from device");
						return false;
					}
#if true
					if (outFlag)
					{
						// Compare what we sent with what we got back
						for (int j = 0; j < bytesToSend; j++)
						{
							if (Device.InputReport[j + 1] != buf[offset + j])
							{
								// TBD: do something with a data mismatch
								return false;
							}
						}
					}
#endif
				}

				bytesLeft -= outLen;
			}

			return Device.DeviceDetected;
		}

		// Send an array of bytes to the device using SET_FEATURE.  They are send
		// in chunks of either len or the length of the output report, whichever is smaller.
		bool SendFeatureBytes(byte reportID, byte[] buf, int len, bool outFlag, bool inFlag)
		{
			bool Result = false;
			int outLen = Device.FeatureDataLength - 1; // Subtract 1 for report ID
			int numSends = (len + outLen - 1) / outLen;
			int bytesLeft = len;
			int offset = 0;
			for (int i = 0; i < numSends; i++, offset += outLen)
			{
				int bytesToSend;
				if (bytesLeft < outLen)
					bytesToSend = bytesLeft;
				else
					bytesToSend = outLen;
				Array.Clear(Device.FeatureReport, 0, Device.FeatureDataLength);
				Device.FeatureReport[0] = reportID; // Report ID

				if (outFlag)
				{
					for (int j = 0; j < bytesToSend; j++)
						Device.FeatureReport[j + 1] = buf[offset + j];

					Result = HIDDevice.HidD_SetFeature(Device.HidDevice, Device.FeatureReport,
						Device.Caps.FeatureReportByteLength);
					if (!Result)
					{
						//The HidD_SetFeature failed, so close the handle, display a message,
						//and set DeviceDetected to false so the next attempt will look for the device.
						Device.CloseHidDevice(true);
						DisplayData("Write failed in SendFeatureBytes()");
						return false;
					}
				}

				if (inFlag)
				{
					// Read the bytes back
					Result = HIDDevice.HidD_GetFeature(Device.HidDevice, Device.FeatureReport, Device.Caps.FeatureReportByteLength);

					if (!Result)
					{
						HIDDevice.HidD_FlushQueue(Device.HidDevice);
						Device.CloseHidDevice(true);
						DisplayData("Read failed in SendFeatureBytes()");
						return false;
					}
#if true
					if (outFlag)
					{
						// Compare what we sent with what we got back (skipping the first
						// byte after the report ID because it contains PORTA data).
						for (int j = 1; j < bytesToSend; j++)
						{
							if (Device.FeatureReport[j + 1] != buf[offset + j])
							{
								// TBD: do something with a data mismatch
								return false;
							}
						}
					}
#endif
				}
				bytesLeft -= outLen;
			}

			return Device.DeviceDetected;
		}

		// Test bandwidth of HID device using the feature report.
		void EchoTest1(bool outFlag, bool inFlag)
		{
			if (!outFlag)
			{
				DisplayData("Invalid test - OUT must be enabled");
				return;
			}

			//If the device hasn't been detected already, look for it.
			if (!Device.DeviceDetected)
			{
				if (!Device.FindHID(VendorID, ProductID))
				{
					DisplayData("Failed to find device");
					return;
				}
			}

			int i, i1, q;
			byte[] outBuf = new byte[1024];
			byte[] inBuf = new byte[1024];
			Array.Clear(outBuf, 0, outBuf.Length);
			Array.Clear(inBuf, 0, inBuf.Length);

			int ctr = 0;
			int charCount;

			failCount = 0;
			for (i1 = 4; i1 <= 64; i1 *= 2)
			{
				int startTime = timeGetTime();

				// i1 = # of entries in the packet.
				charCount = 0;
				const int packetCount = RESEND_COUNT;
				for (i = 0; i < packetCount; i++)
				{
					int ctr0 = ctr;
					for (q = 0; q < i1; q++)
						outBuf[q] = (byte)ctr++;

					bool res = SendFeatureBytes(0, outBuf, i1, outFlag, inFlag);
					charCount += i1 + 1; // length of data + report ID
					if (res)
					{
						if (inFlag)
							charCount += i1 + 1;
					}
					else
					{
						failCount++;
						DisplayData("SendFeatureBytes failed");
						return;
					}
				}

				int endTime = timeGetTime();
				if (failCount > 0)
				{
					string errString = string.Format("Failures: {0}", failCount);
					DisplayData(errString);
				}
				double elapsedTime = (endTime - startTime) / 1000.0;
				int outLen = Device.Caps.OutputReportByteLength - 1; // Subtract 1 for report ID
				int numSends = (i1 + outLen - 1) / outLen;
				int dirCnt = (inFlag ? 2 : 1);
				String statString =
					String.Format("Size {0}, {1} sec, {2} total bytes, {3:f3} packets/msec, (approx {4} Baud, {5} w/o overhead)",
						i1, elapsedTime, charCount,
						(double)(dirCnt * RESEND_COUNT * numSends) / (double)(endTime - startTime),
						(int)(10.0 * charCount / elapsedTime),
						(int)(10.0 * i1 * dirCnt * RESEND_COUNT / elapsedTime));
				DisplayData(statString);
			}
		}


		// Send packets to the programmer and have it echo them back.  Good test of throughput
		// on the serial port.  Overhead ranges from 100% for 4 byte packets to about 10% for
		// the 64 byte maximum size packets.
		void EchoTest2(bool outFlag, bool inFlag)
		{
			if (!outFlag)
			{
				DisplayData("Invalid test - OUT must be enabled");
				return;
			}

			//If the device hasn't been detected already, look for it.
			if (!Device.DeviceDetected)
			{
				if (!Device.FindHID(VendorID, ProductID))
				{
					DisplayData("Failed to find device");
					return;
				}
			}

			int i, i1, q;
			byte[] outBuf = new byte[128];
			byte[] inBuf = new byte[128];
			Array.Clear(outBuf, 0, outBuf.Length);
			Array.Clear(inBuf, 0, inBuf.Length);

			int ctr = 0;
			int charCount;

			if (inFlag)
			{
				// Ensure we clean up input buffers before running the test
				FlushPendingInputReports();
			}

			failCount = 0;
			for (i1 = 4; i1 <= 64; i1 *= 2)
			{
				int startTime = timeGetTime();

				// i1 = # of entries in the packet.
				charCount = 0;
				const int packetCount = RESEND_COUNT;
				for (i = 0; i < packetCount; i++)
				{
					int ctr0 = ctr;
					for (q = 0; q < i1; q++)
						outBuf[q] = (byte)ctr++;

					bool res = SendReportBytes(0, outBuf, i1, outFlag, inFlag);
					charCount += i1 + 1; // length of data + report ID
					if (res)
					{
						if (inFlag)
							charCount += i1 + 1;
					}
					else
					{
						failCount++;
					}
				}

				int endTime = timeGetTime();
				if (failCount > 0)
				{
					string errString = string.Format("Failures: {0}", failCount);
					DisplayData(errString);
				}
				double elapsedTime = (endTime - startTime) / 1000.0;
				int outLen = Device.Caps.OutputReportByteLength - 1; // Subtract 1 for report ID
				int numSends = (i1 + outLen - 1) / outLen;
				int dirCnt = (inFlag ? 2 : 1);
				String statString =
					String.Format("Size {0}, {1} sec, {2} total bytes, {3:f3} packets/msec, (approx {4} Baud, {5} w/o overhead)",
						i1, elapsedTime, charCount,
						(double)(dirCnt * packetCount * numSends) / (double)(endTime - startTime),
						(int)(10.0 * charCount / elapsedTime),
						(int)(10.0 * i1 * dirCnt * RESEND_COUNT / elapsedTime));
				DisplayData(statString);
			}
		}

		void DisplayInputReport()
		{
			int ByteNumber;
			byte ReceivedByte;

			textBox2.AppendText("Input Report: ");
			//Step through the received bytes and display each.
			for (ByteNumber = 0; ByteNumber < Device.Caps.InputReportByteLength; ByteNumber++)
			{
				//Get a byte.
				ReceivedByte = Device.InputReport[ByteNumber];
				//Display it.
				DisplayReceivedData(ReceivedByte);
			}
			textBox2.AppendText("   ");
			string chStr = "";
			for (ByteNumber = 1; ByteNumber < Device.Caps.InputReportByteLength; ByteNumber++)
				chStr += ((char)Device.InputReport[ByteNumber]).ToString();
			textBox2.AppendText(chStr);
			textBox2.AppendText("\r\n");
		}


		void DisplayFeatureReport()
		{
			int ByteNumber;
			byte ReceivedByte;

			textBox2.AppendText("Feature Report: ");
			//Step through the received bytes and display each.
			for (ByteNumber = 0; ByteNumber < Device.Caps.FeatureReportByteLength; ByteNumber++)
			{
				//Get a byte.
				ReceivedByte = Device.FeatureReport[ByteNumber];
				//Display it.
				DisplayReceivedData(ReceivedByte);
			}
			textBox2.AppendText("   ");
			string chStr = "";
			for (ByteNumber = 1; ByteNumber < Device.Caps.FeatureReportByteLength; ByteNumber++)
				chStr += (char)Device.FeatureReport[ByteNumber];
			textBox2.AppendText(chStr);
			textBox2.AppendText("\r\n");
		}

		void DisplayData(string dataToDisplay)
		{
			//Display data in the log List Box
			textBox2.AppendText(dataToDisplay);
			textBox2.AppendText("\r\n");
		}

		void DisplayReceivedData(byte ReceivedByte)
		{
			//Display data received from the device.
			string strByteRead = String.Format("{0:X2} ", ReceivedByte);
			textBox2.AppendText(strByteRead);
		}

		// Send a SET_FEATURE to the device
		bool SetFeature(byte reportID)
		{
			if (!Device.DeviceDetected)
			{
				if (!Device.FindHID(VendorID, ProductID))
					return false;
			}

			// Ensure the feature report is clear before sending.
			Array.Clear(Device.FeatureReport, 0, Device.FeatureReport.Length);
			Device.FeatureReport[0] = reportID;
			// Accumulate bits from the checkboxes (device will use them
			// to do something like turn on/off LEDs).
			Device.FeatureReport[1] = 0;
			if (checkBox1.Checked)
				Device.FeatureReport[1] |= 0x01;
			if (checkBox2.Checked)
				Device.FeatureReport[1] |= 0x02;
			if (checkBox3.Checked)
				Device.FeatureReport[1] |= 0x04;
			if (checkBox4.Checked)
				Device.FeatureReport[1] |= 0x08;
			if (checkBox5.Checked)
				Device.FeatureReport[1] |= 0x10;
			if (checkBox6.Checked)
				Device.FeatureReport[1] |= 0x20;
			if (checkBox7.Checked)
				Device.FeatureReport[1] |= 0x40;
			if (checkBox8.Checked)
				Device.FeatureReport[1] |= 0x80;
			Device.FeatureReport[2] = 0x42; // Data byte 1
			bool Result = HIDDevice.HidD_SetFeature(Device.HidDevice, Device.FeatureReport,
				Device.Caps.FeatureReportByteLength);
			return Result;
		}

		void ReadAndWriteToDevice()
		{
			//If the device hasn't been detected already, look for it.
			if (!Device.DeviceDetected)
			{
				if (Device.FindHID(VendorID, ProductID))
				{
					DisplayData("Device detected");
				}
				else
					DisplayData("Device not detected");
			}

			if (Device.DeviceDetected)
			{
				HIDDevice.HidD_FlushQueue(Device.HidDevice);

				//Write a report to the device.
				if (WriteReport())
				{
					//Read a report from the device.
					if (GetReportBytes())
					{
						//Display the report data.
						DisplayInputReport();
					}
				}

#if true
				SetFeature(0);
				Array.Clear(Device.FeatureReport, 0, Device.FeatureReport.Length);
				bool hRes = HIDDevice.HidD_GetFeature(Device.HidDevice, Device.FeatureReport, Device.Caps.FeatureReportByteLength);
				if (hRes)
				{
					//Display the report data.
					DisplayFeatureReport();
				}
#endif
			}
		}
		private void toolStripButton1_Click(object sender, EventArgs e)
		{
			ReadAndWriteToDevice();
		}

		private void checkBox1_CheckedChanged(object sender, EventArgs e)
		{
			SetFeature(0);
		}

		private void toolStripButton2_Click(object sender, EventArgs e)
		{
			EchoTest2(true, false);
		}

		private void toolStripButton3_Click(object sender, EventArgs e)
		{
			EchoTest1(true, false);
		}

		private void toolStripButton4_Click(object sender, EventArgs e)
		{
			EchoTest2(true, true);
		}

		private void toolStripButton5_Click(object sender, EventArgs e)
		{
			EchoTest1(true, true);
		}
	}
}
